import time
import threading

import psycopg2
import redis

from zope.interface import implements
from twisted.internet import reactor
from twisted.web import resource
from txstatsd.itxstatsd import IMetric

ONE_MINUTE = 60
ONE_HOUR = 60 * ONE_MINUTE
ONE_DAY = 24 * ONE_HOUR


class DistinctResource(resource.Resource):
    """"""
    isLeaf = True

    def __init__(self, reporter):
        resource.Resource.__init__(self)
        self.reporter = reporter

    def render_GET(self, request):
        """"""
        pass

    def getDistinctCount(self, since, until):
        """"""
        c = psycopg2.connect(self.dsn)
        cr = c.cursor()
        cr.execute("SELECT * FROM paths WHERE path = %s", (self.path,))

    def getTop(self, since, until, count):
        """"""
        pass

class DistinctMetricReporter(object):
    """
    Keeps an mesurement of the distinct numbers of items seen and the times
    it has seen each one.
    """
    implements(IMetric)

    periods = [5 * ONE_MINUTE, ONE_HOUR, ONE_DAY]

    def __init__(self, name, wall_time_func=time.time, prefix="",
            bucket_size=ONE_DAY, dsn=None, redis_host=None, redis_port=None):
        """Construct a metric we expect to be periodically updated.

        @param name: Indicates what is being instrumented.
        @param wall_time_func: Function for obtaining wall time.
        @param prefix: If present, a string to prepend to the message
            composed when C{report} is called.
        """
        self.name = name
        self.wall_time_func = wall_time_func
        if prefix:
            prefix += '.'
        self.prefix = prefix
        self.bucket_size = bucket_size
        self.dsn = dsn
        self.redis_host = redis_host
        if redis_port is None:
            redis.port = 6379
        self.redis_port = redis_port
        self.metric_id = None
        self.build_bucket()
        self.redis_flush_lock = threading.Lock()
        self.redis_count = {}

        if redis_host != None:
            self.redis = redis.client.Redis(host=redis_host, port=redis_port)

    def build_bucket(self, timestamp=None):
        self.max = 0
        self.bucket = {}
        self.bucket_no = self.get_bucket_no(timestamp)

    def get_bucket_no(self, timestamp=None):
        if timestamp is None:
            timestamp = self.wall_time_func()
        return int(timestamp / (self.bucket_size))

    def process(self, fields):
        self.update(fields[0])

    def update(self, item):
        value = self.bucket.get(item, 0) + 1

        self.bucket[item] = value
        if value > self.max:
            self.max = value

        now = self.wall_time_func()
        if self.redis_host is not None:
            reactor.callInThread(self._update_count, item, now)

    def bucket_name_for(self, period):
        return "bucket_" + str(period)

    def _update_count(self, value, when):
        for period in self.periods:
            self.redis.zadd(self.bucket_name_for(period), value, when)

    def _flush_redis(self, now):
        if self.redis_flush_lock.acquire(False) is False:
            return
        try:
            for period in self.periods:
                bucket = self.bucket_name_for(period)
                self.redis.zremrangebyscore(bucket, 0, now - period)
                self.redis_count[bucket] = self.redis.zcard(bucket)
        finally:
            self.redis_flush_lock.release()

    def count(self, period):
        return self.redis_count.get(self.bucket_name_for(period), 0)

    def count_5min(self):
        return self.count(5 * ONE_MINUTE)

    def count_1hour(self):
        return self.count(ONE_HOUR)

    def count_1day(self):
        return self.count(ONE_DAY)

    def _save_bucket(self, bucket, bucket_no):
        path = self.prefix + self.name
        if self.metric_id is None:
            c = psycopg2.connect(self.dsn)
            cr = c.cursor()
            cr.execute("SELECT * FROM paths WHERE path = %s", (path,))
            row = cr.fetchone()
            if row is None:
                cr.execute("INSERT INTO paths (path) VALUES (%s) "
                            "RETURNING (id)", (path,))
                row = cr.fetchone()
                cr.execute("commit")

            self.metric_id = row[0]

        for i, (k, v) in enumerate(bucket.iteritems()):
            cr.execute("INSERT INTO points (path_id, bucket, value, count) "
                "VALUES (%s, %s, %s, %s)", (self.metric_id, bucket_no,
                                            k, v))
            if i % 1000 == 0:
                cr.execute("commit")
        cr.execute("commit")

    def save_bucket(self, bucket, bucket_no):
        if self.dsn is not None:
            reactor.callInThread(self._save_bucket, bucket, bucket_no)

    def flush(self, interval, timestamp):
        current_bucket = self.get_bucket_no(timestamp)
        if current_bucket != self.bucket_no:
            self.save_bucket(self.bucket, self.bucket_no)
            self.build_bucket(timestamp)

        if self.redis_host is not None:
            reactor.callInThread(self._flush_redis, timestamp)

        metrics = []
        items = {".messages": len(self.bucket),
                 ".max": self.max,
                 ".count_5min": self.count_5min(),
                 ".count_1hour": self.count_1hour(),
                 ".count_1day": self.count_1day(),
                 }
        for item, value in items.iteritems():
            metrics.append((self.prefix + self.name + item, value, timestamp))
        return metrics
