# Copyright (C) 2011 Canonical
# All Rights Reserved

import ConfigParser
from cStringIO import StringIO
import os
import time
try:
    from subprocess import check_output
except ImportError:
    import subprocess
    def check_output(args):
        return subprocess.Popen(args,
            stdout=subprocess.PIPE).communicate()[0]

import psycopg2
import redis

from twisted.trial.unittest import TestCase
from twisted.internet import reactor
from twisted.plugin import getPlugins
from twisted.plugins import distinctdbplugin
from txstatsd.itxstatsd import IMetricFactory
from txstatsd import service

from distinctdb import distinctmetric as distinct


class TestDistinctMetricReporter(TestCase):

    def test_get_bucket_no(self):
        _wall_time = [0]

        def _time():
            return _wall_time[0]

        dmr = distinct.DistinctMetricReporter("test", wall_time_func=_time)
        self.assertEquals(dmr.get_bucket_no(), 0)
        _wall_time = [60 * 60 * 24 + 1]
        dmr.update("one")
        self.assertEquals(dmr.get_bucket_no(), 1)

    def test_max(self):
        _wall_time = [0]

        def _time():
            return _wall_time[0]

        dmr = distinct.DistinctMetricReporter("test", wall_time_func=_time)
        self.assertEquals(dmr.get_bucket_no(), 0)
        self.assertEquals(dmr.max, 0)
        dmr.update("one")
        dmr.update("one")
        dmr.update("two")
        self.assertEquals(dmr.max, 2)
        dmr.flush(1, 60 * 60 * 24 + 1)
        dmr.update("one")
        self.assertEquals(dmr.max, 1)

    def test_reports(self):
        _wall_time = [0]

        def _time():
            return _wall_time[0]

        result = {}

        dmr = distinct.DistinctMetricReporter("test", wall_time_func=_time)

        def save(b, b_no):
            result["bucket"] = b
            result["bucket_no"] = b_no
        dmr.save_bucket = save
        dmr.update("one")
        dmr.update("one")
        dmr.update("two")
        day = 60 * 60 * 24 + 1
        dmr.flush(1, day)
        dmr.update("three")
        self.assertEquals(result,
            {"bucket": {"one": 2, "two": 1}, "bucket_no": 0})
        result = dmr.flush(1, day)

        self.assertTrue(("test.max", 1, day) in result)
        self.assertTrue(("test.messages", 1, day) in result)

    def test_configure(self):
        class TestOptions(service.OptionsGlue):
            optParameters = [["test", "t", "default", "help"]]
            config_section = "statsd"

        o = TestOptions()
        config_file = ConfigParser.RawConfigParser()
        config_file.readfp(StringIO("[statsd]\n\n[plugin_distinctdb]\n"
            "dsn = dbdsn\nbucket_size = 100"))
        o.configure(config_file)
        dmf = distinctdbplugin.DistinctMetricFactory()
        dmf.configure(o)
        dmr = dmf.build_metric("foo", "bar", time.time)
        self.assertEquals(dmr.bucket_size, 100)
        self.assertEquals(dmr.dsn, "dbdsn")


class TestPlugin(TestCase):

    def test_factory(self):
        self.assertTrue(distinctdbplugin.distinct_metric_factory in \
                        list(getPlugins(IMetricFactory)))


class TestDatabase(TestCase):

    def setUp(self):
        rootdir = check_output(["bzr", "root"]).strip()
        dsn_file = os.path.join(rootdir, "tmp", "pg.dsn")
        self.dsn = open(dsn_file).read()
        self.conn = psycopg2.connect(self.dsn)

    def tearDown(self):
        cr = self.conn.cursor()
        cr.execute("rollback")
        cr.execute("DELETE FROM paths")
        cr.execute("DELETE FROM points")
        cr.execute("commit")

    def test_connect(self):
        cr = self.conn.cursor()
        cr.execute("SELECT 0")
        result = cr.fetchall()
        self.assertTrue(result, [(0,)])

    def test_create_metric_id(self):
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr._save_bucket({}, 0)
        cr = self.conn.cursor()
        cr.execute("SELECT * FROM paths WHERE path = 'test'")
        cr.execute("SELECT * FROM paths")
        self.assertEquals(len(cr.fetchall()), 1)

    def test_find_saved_data(self):
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr.update("one")
        dmr.update("one")
        dmr.update("two")
        dmr._save_bucket(dmr.bucket, 0)
        cr = self.conn.cursor()
        cr.execute("SELECT * FROM points ORDER BY value")
        rows = cr.fetchall()
        self.assertEquals(rows, [(dmr.metric_id, 0, "one", 2),
                                 (dmr.metric_id, 0, "two", 1)])

    def test_load_metric_id(self):
        dmr = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr._save_bucket({}, 0)
        dmr2 = distinct.DistinctMetricReporter("test", dsn=self.dsn)
        dmr2._save_bucket({}, 0)
        self.assertEquals(dmr.metric_id, dmr2.metric_id)


class TestRedis(TestCase):

    def setUp(self):
        reactor._initThreadPool()
        reactor.threadpool.start()

    def tearDown(self):
        r = redis.client.Redis(host="localhost", port=16379)
        r.flushdb()
        reactor.threadpool.stop()

    def test_connect(self):
        r = redis.client.Redis(host="localhost", port=16379)
        r.ping()

    def test_configure(self):
        class TestOptions(service.OptionsGlue):
            optParameters = [["test", "t", "default", "help"]]
            config_section = "statsd"

        o = TestOptions()
        config_file = ConfigParser.RawConfigParser()
        config_file.readfp(StringIO("[statsd]\n\n[plugin_distinctdb]\n"
            "redis_host = localhost\nredis_port = 16379"))
        o.configure(config_file)
        dmf = distinctdbplugin.DistinctMetricFactory()
        dmf.configure(o)
        dmr = dmf.build_metric("foo", "bar", time.time)
        self.assertEquals(dmr.redis_host, "localhost")
        self.assertEquals(dmr.redis_port, 16379)

    def test_usage(self):
        dmr = distinct.DistinctMetricReporter("test",
            redis_host="localhost", redis_port=16379)

        self.assertEquals(dmr.count_1hour(), 0)
        dmr._update_count("one", 0)
        dmr._flush_redis(1)
        self.assertEquals(dmr.count_1hour(), 1)
        dmr._update_count("one", 0)
        dmr._flush_redis(1)
        self.assertEquals(dmr.count_1hour(), 1)
        dmr._update_count("two", 30 * distinct.ONE_MINUTE)
        dmr._flush_redis(30 * distinct.ONE_MINUTE)
        self.assertEquals(dmr.count_1hour(), 2)
        dmr._flush_redis(distinct.ONE_HOUR + 10 * distinct.ONE_MINUTE)
        self.assertEquals(dmr.count_1hour(), 1)

    def test_load(self):
        dmr = distinct.DistinctMetricReporter("test",
            redis_host="localhost", redis_port=16379)
        start = time.time()
        for i in range(10000):
            dmr.update(str(i % 1000))

        while True:
            w = len(reactor.threadpool.working)
            if w == 0:
                break
            time.sleep(0.1)
        dmr._flush_redis(time.time())
        duration = time.time() - start
        self.assertEquals(dmr.count_1hour(), 1000)
        self.assertTrue(duration <  10)
